import pickle
import importlib.util
import threading as mt
import multiprocessing as mp

from collections.abc import Iterator
from queue import Empty
from traceback import format_tb
from typing import Iterable, Mapping, Callable, Optional, Union, Sequence, Any, Literal

from coba.utilities import peek_first
from coba.exceptions import CobaException
from coba.pipes.primitives import Filter, Line, SourceSink
from coba.pipes.filters import Slice
from coba.pipes.sources import IterableSource, QueueSource
from coba.pipes.sinks import QueueSink

# handle not picklable (this is handled by explicitly pickling) (UNITTESTED)
# handle empty list (this is done  naturally) (UNITTESTED)
# handle exceptions in process (wrap worker executing code in an exception handler) (UNITTESTED)
# handle AttributeErrors. This occurs when... (handled by Pickler) (UNITTESTED)

# handle ctrl-c without hanging (manually tested)
#   > This is super hard... I've been trying to do this for years... here are things I've done
#   > I make all threads and processes daemons to protect against them not closing
#   > When I'm done I shutdown in-queue loading and then empty both the in-queue and out-queue
#   > I use try-finally idioms to close all queues that we create
#   > I assert that all processes are in fact closed when passed to the callback

# handle Experiment.evaluate not being called inside of __name__=='__main__' (manually tested)
#   > Handled by checking exitcode on processes and using an event synchronizer on first process start

# There are three potential contexts -- spawn, fork, and forkserver. On Windows and Mac spawn is the only option.
# On Linux the default option is fork. Fork creates processes faster and can share memory but also doesn't play
# well with threads. Therefore, to make behavior consistent across Linux, Windows, and Mac -- and to avoid
# potential bugs -- we force our multiprocessors to always use spawn regardless of the host OS.
spawn_context = mp.get_context("spawn")

class ProcessLine(spawn_context.Process):

    ### We create a lock so that we can safely receive any possible exceptions. Empirical
    ### tests showed that creating a Pipe and Lock doesn't seem to slow us down too much.

    def __init__(self, line: Line, callback: Callable = None):

        self._line      = line
        self._callback  = callback
        self._exception = None
        self._traceback = None
        self._poisoned  = False
        super().__init__(daemon=True)

    def start(self) -> None:
        callback               = self._callback
        self._callback         = None
        self._recv, self._send = spawn_context.Pipe(False)
        self._lock             = spawn_context.Lock()

        super().start()

        if callback:
            def join_and_call(self=self):
                self.join()
                callback(self)
            mt.Thread(target=join_and_call,daemon=True).start()

    def run(self): #pragma: no cover (coverage can't be tracked for code that runs on background prcesses)
        try:
            self._line.run()
        except Exception as e:
            if str(e).startswith("Can't get attribute"):
                ex,tb = CobaException(
                    "We attempted to evaluate your code in multiple processes but we were unable to find all the code "
                    "definitions needed to pass the tasks to the processes. The two most common causes of this error are: "
                    "1) a learner or simulation is defined in a Jupyter Notebook cell or 2) a necessary class definition "
                    "exists inside the `__name__=='__main__'` code block in the main execution script. In either case you "
                    "can choose one of four simple solutions: 1) pip install cloudpickle into your python environment, 2) "
                    "evaluate your code on a single process, 3) if in Jupyter notebook define all necessary classes in a "
                    "separate file and include the classes via import statements, or 4) move your class definitions outside "
                    "the `__name__ == '__main__'` check."
                ),None
            else:
                ex,tb = e,format_tb(e.__traceback__)
        except KeyboardInterrupt as e:
            ex,tb = e,None
        else:
            ex,tb = None,None

        self._send.send((ex, tb, hasattr(self._line[0],'_poisoned') and self._line[0]._poisoned))

    def join(self) -> None:
        super().join()
        self._get_result()

    @property
    def pipeline(self) -> Line:
        return self._line

    @property
    def traceback(self) -> Sequence[str]:
        return self._traceback

    @property
    def exception(self) -> Optional[Exception]:
        return self._exception

    @property
    def poisoned(self) -> bool:
        return self._poisoned

    def _get_result(self):
        with self._lock:
            if not self._recv.closed:
                if self._recv.poll():
                    ex,tb,po = self._recv.recv()
                    self._exception = ex
                    self._traceback = tb
                    self._poisoned  = po
                self._send.close()
                self._recv.close()

class ThreadLine(mt.Thread):

    def __init__(self, line: Line, callback: Callable = None) -> None:
        self._line      = line
        self._callback  = callback
        self._exception = None
        self._traceback = None
        self._poisoned  = False
        super().__init__(daemon=True)

    def start(self) -> None:

        super().start()

        if self._callback:
            #we start a thread and join so that the callback
            #is not called until the thread is no longer alive
            def join_and_call():
                self.join()
                self._callback(self)
            mt.Thread(target=join_and_call,daemon=True).start()

    def run(self) -> None:
        try:
            self._line.run()
        except Exception as e:
            self._exception = e
            self._traceback = format_tb(e.__traceback__)
        self._poisoned  = hasattr(self._line[0],'_poisoned') and self._line[0]._poisoned

    @property
    def pipeline(self) -> Line:
        return self._line

    @property
    def traceback(self) -> Sequence[str]:
        return self._traceback

    @property
    def exception(self) -> Optional[Exception]:
        return self._exception

    @property
    def poisoned(self) -> bool:
        return self._poisoned

class AsyncableLine(SourceSink):

    def run_async(self,
        callback:Callable[['AsyncableLine',Optional[Exception],Optional[str]],None]=None,
        mode:Literal['process','thread']='process') -> Union[ThreadLine,ProcessLine]:
        """Run the pipeline asynchronously."""

        mode = mode.lower()

        if mode == "process":
            worker = ProcessLine(line=self, callback=callback)
        elif mode == "thread":
            worker = ThreadLine(line=self, callback=callback)
        else:
            raise CobaException(f"Unrecognized pipe async mode {mode}. Valid values are 'process' and 'thread'.")

        worker.start()
        return worker

class Safe:
    def __init__(self, filter: Filter):
        if importlib.util.find_spec('cloudpickle'):
            import cloudpickle
            try:
                self._filter = cloudpickle.dumps(filter)
            except:
                self._filter = filter
        else:
            self._filter = filter

    def filter(self,items):
        if importlib.util.find_spec('cloudpickle') and isinstance(self._filter,bytes):
            import cloudpickle
            filter = cloudpickle.loads(self._filter)
        else:
            filter = self._filter

        yield from filter.filter(items)

class Foreach:
    def __init__(self,pipe:Filter) -> None:
        self._pipe = pipe

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:
        for item in items:
            out = self._pipe.filter(item)
            out = out if isinstance(out,Iterator) else [out]
            yield from out

class Pickler:
    def filter(self, items) -> Iterable[bytes]:
        try:
            if importlib.util.find_spec('cloudpickle'):
                import cloudpickle
                yield from map(cloudpickle.dumps,items)
            else:
                yield from map(pickle.dumps,items)
        except Exception as e:
            if "pickle" in str(e) or "Pickling" in str(e):
                message = (
                    f"We attempted to process your code on multiple processes but were unable to do so due to a pickle "
                    f"error. The exact error received was '{str(e)}'. Errors of this kind can often be fixed in one of three "
                    f"ways: 1) pip install cloudpickle in your conda environment, 2) evaluate the experiment in question on "
                    f"a single process with no limit on the tasks per child or 3) modify the named class to be picklable. "
                    f"The easiest way to make a given class picklable is to add `def __reduce__(self): return (<the class "
                    f"in question>, (<tuple of constructor arguments>))` to the class. For more information see "
                    f"https://docs.python.org/3/library/pickle.html#object.__reduce__."
                )
                raise CobaException(message)
            else: #pragma: no cover
                raise

class Unpickler:
    def filter(self, items):
        if importlib.util.find_spec('cloudpickle'):
            import cloudpickle
            yield from map(cloudpickle.loads,items)
        else:
            yield from map(pickle.loads,items)

class EventSetter:
    def __init__(self, event: mt.Event) -> None:
        self._event = event
    def filter(self, items):
        self._event.set()
        return items

class Stopper:
    def __init__(self) -> None:
        self._stop = False

    def stop(self) -> None:
        self._stop = True

    def filter(self, items):
        for item in items:
            if self._stop: break
            yield item

class Multiprocessor(Filter[Iterable[Any], Iterable[Any]]):
    """Create multiple processes to filter given items."""

    def __init__(self,
            filter: Filter[Iterable[Any], Iterable[Any]],
            n_processes: int = 1,
            maxtasksperchild: int = 0) -> None:
        """Instantiate a Multiprocessor.

        Args:
            filter: The inner pipe that will be executed on multiple processes.
            n_processes: The number of processes that should be created to filter items.
            maxtasksperchild: The number of items/chunks a process should filter before restarting.
        """
        self._filter           = filter
        self._max_processes    = n_processes
        self._maxtasksperchild = maxtasksperchild or None

    @property
    def params(self) -> Mapping[str,Any]:
        return self._filter.params

    def filter(self, items: Iterable[Any]) -> Iterable[Any]:

        items = peek_first(items)[1]
        if not items: return []

        if self._max_processes == 1 and self._maxtasksperchild is None:
            yield from Foreach(self._filter).filter(items)

        else:
            event = spawn_context.Event()

            #for some reason if this mp queue get too big we can't keyboradinterrupt
            #therefore, we slightly limit its size and then empty it before closing.
            in_queue  = spawn_context.Queue(maxsize=self._max_processes*2)
            out_queue = spawn_context.Queue()
            in_put    = QueueSink(in_queue,foreach=True)
            in_get    = QueueSource(in_queue)
            out_put   = QueueSink(out_queue,foreach=True)
            out_get   = QueueSource(out_queue)
            pickler   = Pickler()
            unpickler = Unpickler()
            get_max   = Slice(None,self._maxtasksperchild)
            setter    = EventSetter(event)

            self._n_procs      = self._max_processes
            self._exceptions   = []
            self._poison       = None
            self._main_err     = False
            self._load_stopper = Stopper() #this works because the loader is a thread which means we have shared memory

            load_line   = SourceSink(IterableSource(items), self._load_stopper, pickler, in_put)
            filter_line = SourceSink(in_get, setter, unpickler, get_max, Safe(Foreach(self._filter)), out_put)

            def loader_finished_or_failed(worker: Union[ThreadLine,ProcessLine]):
                if worker.exception: self._exceptions.append(worker.exception)
                in_put.write(self._load_stopper.filter([self._poison]*self._n_procs))

            def filter_finished_or_failed(worker: Union[ThreadLine,ProcessLine]):
                if worker.exception: self._exceptions.append(worker.exception)

                assert not worker.is_alive()

                #only known cause of exitcode != 0 is a missing `if __name__ == '__main__'``.
                if worker.exitcode != 0: #pragma: no cover
                    #exitcode -15 is keyboard interrupt...
                    if worker.exitcode != -15:
                        print(worker.exitcode)
                    self._main_err = True
                    event.set()

                #we have to stop on exception since, depending on where the exception occurred,
                #we may not have actually read anything from the input queue. If we didn't then
                #the input queue will never empty and we'll be stuck starting processes forever.
                if not worker.poisoned and not self._exceptions and worker.exitcode == 0:
                    ProcessLine(worker.pipeline,callback=filter_finished_or_failed).start()
                else:
                    self._n_procs -= 1
                    if self._n_procs == 0:
                        try:
                            out_put.write([self._poison])
                        except ValueError: #pragma: no cover
                            pass

            load_thread = ThreadLine(load_line, callback=loader_finished_or_failed)
            filt_procs  = [ProcessLine(filter_line, callback=filter_finished_or_failed) for _ in range(self._n_procs)]

            try:
                load_thread.start()
                filt_procs.pop().start()

                #by waiting we can avoid throwing multiple exceptions
                #when there is a problem with starting a new process
                event.wait()
                if not self._main_err:
                    for p in filt_procs: p.start()
                    yield from out_get.read()

            finally:

                #stop loading into the input queue
                self._load_stopper.stop()

                #empty the input queue and then close it
                #if we don't empty first then we can easily
                #lock during a keyboard interrupt
                try:
                    while True: in_queue.get_nowait()
                except Empty:
                    pass
                    #closing can cause exceptions
                    #and doesn't seem to help anything
                    #in_queue.close()

                #empty the input queue and then close it
                #if we don't empty first then we can easily
                #lock during a keyboard interrupt
                try:
                    while True: out_queue.get_nowait()
                except Empty:
                    pass
                    #closing can cause exceptions
                    #and doesn't seem to help anything
                    #out_queue.close()

            if self._exceptions:
                raise self._exceptions[0]



"""A custom implementation of random number generation.

This module follows the pattern of the standard library's random module by creating
by instantiation an internal, global Random class and then referencing that in all
the public methods in to maintain state.

Remarks:
    This implementation has been made to guarantee the reproducibility of experiments
    according to integer seeds across all versions of Python. The standard implementation
    of random within Python has had a few variations in implementation in the past and
    could always change in the future, making randomization by seed potentially non-fixed.
"""

import math
import time

from itertools import compress, accumulate, repeat, islice
from operator import le
from typing import Optional, Iterable, Sequence, Any

class CobaRandom:
    """A random number generator that is consistent across python implementations."""

    def __init__(self, seed: Optional[float] = None) -> None:
        """Instantiate a CobaRandom.

        Args:
            seed: The seed to use when starting random number generation.

        Remarks:
            The values for a,c,m below are taken from L'ecuyer (1999). In this paper he notes that
            these values for LCG should have an overall period of m though the period of lower bits
            will be much shorter. A solution he offers to this problem is to use much larger m (e.g.,
            2**128) and then only use the the top n most significant digits. For now, we aren't doing that.

        References:
            L'ecuyer, Pierre. "Tables of linear congruential generators of different sizes
            and good lattice structure." Mathematics of Computation 68.225 (1999): 249-260.
        """

        if isinstance(seed,int) or (isinstance(seed,float) and seed.is_integer()):
            seed = int(seed)
        else:
            seed = int.from_bytes(str(seed or time.time()).encode('utf-8'),"big") % 2**20

        self._seed = seed
        self._randu = self._next_uniform(116646453,seed,9,2**30)
        self._randg = self._next_gaussian()

    def randoms(self, n:int, min:float=0, max:float=1) -> Sequence[float]:
        """Generate `n` uniform random numbers in [`min`,`max`).

        Args:
            n: How many random numbers should be generated.
            min: The minimum value for the random numbers.
            max: The maximum value for the random numbers.

        Returns:
            The `n` generated random numbers in [`min`,`max`].
        """
        if (n is not None) and (n < 0 or not isinstance(n, int)):
            raise ValueError("n must be an integer greater than or equal to 0")

        if min == 0 and max == 1:
            iterable = islice(self._randu,n)
        elif min == 0:
            iterable = ( max*r for r in islice(self._randu,n) )
        else:
            r_range = max-min
            iterable = (min+r_range*r for r in islice(self._randu,n))
        
        return list(iterable) if n is not None else iterable

    def random(self, min:float=0, max:float=1) -> float:
        """Generate a uniform random number in [`min`,`max`].

        Args:
            min: The minimum value for the random numbers.
            max: The maximum value for the random numbers.

        Returns:
            The generated random number in [`min`,`max`].
        """
        return min+(max-min)*next(self._randu)        

    def shuffle(self, sequence: Sequence[Any], inplace: bool = False) -> Sequence[Any]:
        """Shuffle the order of items in a sequence.

        Args:
            sequence: The sequence of items that are to be shuffled.
            inplace: The sequence of items should be shuffled within the given sequence

        Returns:
            A new sequence with the order of items shuffled.

        Remarks:
            This is the Richard Durstenfeld's method popularized by Donald Knuth in The Art of Computer
            Programming. This algorithm is unbiased (i.e., all possible permutations are equally likely to occur).
        """

        n = len(sequence)
        if n < 2: return sequence

        l = sequence if inplace else list(sequence)

        #i goes from 0 to n-2
        #j is always i <= j < n
        for i,r in islice(enumerate(self._randu),n-1):
            j = i+int(r*(n-i)) 
            l[i], l[j] = l[j], l[i]
        
        return l

    def randint(self, a:int, b:int) -> int:
        """Generate a uniform random integer in [a, b].

        Args:
            a: The inclusive lower bound for the random integer.
            b: The inclusive upper bound for the random integer.
        """

        return a+int((b-a+1)*next(self._randu))

    def randints(self, n:int, a:int, b:int) -> Sequence[int]:
        """Generate `n` uniform random integers in [a, b].

        Args:
            n: The number of random integers to generate.
            a: The inclusive lower bound for the random integer.
            b: The inclusive upper bound for the random integer.
        """
        b=b+1
        if a == 0:
            return [int(b*r) for r in islice(self._randu,n)]
        else:
            r_range = b-a
            return [int(r_range*r) + a for r in islice(self._randu,n)]

    def choice(self, seq: Sequence[Any], weights:Sequence[float] = None) -> Any:
        """Choose a random item from the given sequence.

        Args:
            seq: The sequence to pick randomly from.
            weights: The proportion by which seq is selected from.
        """

        if weights is None:
            return seq[int(len(seq)*next(self._randu))]

        else:
            tot = sum(weights)
            if tot == 0: raise ValueError("The sum of weights cannot be zero.")
            cdf = accumulate(weights)
            rng = next(self._randu) * tot

            return next(compress(seq, map(le, repeat(rng), cdf)))

    def gauss(self, mu:float=0, sigma:float=1) -> float:
        """Generate a random number from N(mu,sigma).

        Args:
            mu: The expectation of the distribution we are drawing from.
            sigma: The standard deviation of the distribution we are drawing form.

        Returns:
            A random number drawn from N(mu,sigma).
        """
        return self.gausses(1, mu, sigma)[0]

    def gausses(self, n:int, mu:float=0, sigma:float=1) -> Sequence[float]:
        """Generate `n` independent random numbers from N(mu,sigma).

        Args:
            n: The number of random numbers to generate.
            mu: The expectation of the distribution we are drawing from.
            sigma: The standard deviation of the distribution we are drawing form.

        Returns:
            The `n` random numbers drawn from N(mu,sigma).
        """
        return [mu+sigma*g for g in islice(self._randg,n) ]

    def _next_uniform(self, a, s, c, m) -> Iterable[float]:
        """Generate uniform random numbers in [0,1).

        Random numbers are generated using a linear congruential generator.
        """
        m_1 = m-1
        while True:
            #when m is a power of 2 
            #this is equal to modulo m
            s = (a * s + c) & (m_1)
            yield s/m

    def _next_gaussian(self) -> Iterable[float]:
        """Generate `n` gaussian random numbers in N(0,1).
        
        Random numbers are generated using the Box-Muller transform.
        """

        sqrt = math.sqrt
        log  = math.log
        pi   = math.pi
        cos  = math.cos
        sin  = math.sin

        while True:
            R = sqrt(-2*log(next(self._randu)))
            S = 2*pi*next(self._randu)
            yield R*cos(S)
            yield R*sin(S)

    def __reduce__(self):
        return (CobaRandom,(self._seed,))

_random = CobaRandom()

def seed(seed: Optional[float]) -> None:
    """Set the seed for generating random numbers in this module.

    Args:
        seed: The seed for generating random numbers.

    Remarks:
        Note, this seed does not affect random numbers generated by the standard library
    """

    global _random

    _random = CobaRandom(seed)

def random(min:float=0, max:float=1) -> float:
    """Generate a uniform random number in [`min`,`max`].

    Args:
        min: The minimum value for the random numbers.
        max: The maximum value for the random numbers.

    Returns:
        A uniform random number in [`min`,`max`].
    """
    return _random.random(min,max)

def randoms(n: int, min:float=0, max:float=1) -> Sequence[float]:
    """Generate `n` uniform random numbers in [`min`,`max`].

    Args:
        n: How many uniform random numbers should be generated.
        min: The minimum value for the random numbers.
        max: The maximum value for the random numbers.

    Returns:
        The `n` uniform random numbers in [`min`,`max`].
    """

    return _random.randoms(n,min,max)

def gauss(mu:float=0, sigma:float=1) -> float:
    """Generate a random number from N(mu,sigma).

    Returns:
        A random number drawn from N(mu,sigma).
    """

    return _random.gauss(mu,sigma)

def gausses(n:int, mu:float=0, sigma:float=1) -> Sequence[float]:
    """Generate `n` independent random numbers from N(mu,sigma).

    Args:
        n: How many random numbers should be generated
        mu: The expectation of the distribution we are drawing from.
        sigma: The standard deviation of the distribution we are drawing form.

    Returns:
        The `n` random numbers drawn from N(mu,sigma).
    """

    return _random.gausses(n,mu,sigma)

def randint(a:int, b:int) -> int:
    """Generate a uniform random integer in [a, b].

    Args:
        a: The inclusive lower bound for the random integer.
        b: The inclusive upper bound for the random integer.
    """

    return _random.randint(a,b)

def randints(n:int, a:int, b:int) -> Sequence[int]:
    """Generate `n` uniform random integers in [a, b].

    Args:
        n: The number of random integers to generate.
        a: The inclusive lower bound for the random integer.
        b: The inclusive upper bound for the random integer.
    """

    return _random.randints(n,a,b)

def choice(seq: Sequence[Any], weights:Sequence[float]=None) -> Any:
    """Choose a random item from the given sequence.

    Args:
        seq: The sequence to pick randomly from.
        weights: The proportion by which seq is selected from.
    """

    return _random.choice(seq, weights)

def shuffle(array_like: Sequence[Any], inplace:bool=False) -> Sequence[Any]:
    """Shuffle the order of items in a sequence.

    Args:
        sequence: The sequence of items that are to be shuffled.
        inplace: The sequence of items should be shuffled within the given sequence.

    Returns:
        A new sequence with the order of items shuffled.
    """

    return _random.shuffle(array_like, inplace)



from copy import deepcopy
from itertools import islice
from operator import itemgetter
from collections import defaultdict, Counter
from typing import Any, Iterable, Sequence, Optional, Tuple, Mapping

from coba.learners import Learner, SafeLearner
from coba.environments import Environment, SafeEnvironment, Finalize, BatchSafe, Chunk
from coba.evaluators import Evaluator, SafeEvaluator

from coba.pipes import Source, Filter, SourceFilters
from coba.context import CobaContext
from coba.utilities import peek_first

from coba.experiments.results import Result, Table

class Task:

    def __init__(self,
        env : Optional[Tuple[int,Environment]],
        lrn : Optional[Tuple[int,Learner,bool]],
        val : Optional[Tuple[int,Evaluator]],
        copy: bool = False) -> None:
        (self.env_id, self.env) = env or (None,None)
        (self.lrn_id, self.lrn) = lrn or (None,None)
        (self.val_id, self.val) = val or (None,None)
        self.copy   = copy

    def __eq__(self, o: object) -> bool:
        return isinstance(o,Task) \
        and self.env_id == o.env_id \
        and self.env    == o.env    \
        and self.lrn_id == o.lrn_id \
        and self.lrn    == o.lrn    \
        and self.val_id == o.val_id \
        and self.val    == o.val    \
        and self.copy   == o.copy   \

class MakeTasks(Source[Iterable[Task]]):

    def __init__(self,
        triples: Sequence[Tuple[Environment,Learner,Evaluator]],
        restored: Optional[Result] = None) -> None:

        self._triples = triples
        self._restored = restored or Result()

    def read(self) -> Iterable[Task]:

        #we rely on ids to make sure we don't do duplicate work. So long as self.evaluation_pairs
        #is always in the exact same order we should be fine. In the future we may want to consider.
        #adding a better check for environments other than assigning an index based on their order.

        envs = {None:None}
        lrns = {None:None}
        vals = {None:None}

        restored_lrns = set(self._restored.learners['learner_id'])
        restored_envs = set(self._restored.environments['environment_id'])
        restored_vals = set(self._restored.evaluators['evaluator_id'])
        restored_outs = set(zip(*self._restored.interactions[['environment_id','learner_id','evaluator_id']]))

        learner_counts = Counter([l for _,l,_ in self._triples])

        for env, lrn, val in self._triples:

            if env not in envs:
                envs[env] = len(envs)-1
                if envs[env] not in restored_envs:
                    yield Task((envs[env],env),None,None)

            if lrn not in lrns:
                lrns[lrn] = len(lrns)-1
                if lrns[lrn] not in restored_lrns:
                    yield Task(None,(lrns[lrn],lrn),None)

            if val not in vals:
                vals[val] = len(vals)-1
                if vals[val] not in restored_vals:
                    yield Task(None,None,(vals[val],val))

            eid,lid,vid = (envs[env],lrns[lrn],vals[val])
            if val and (eid,lid,vid) not in restored_outs:
                yield Task((eid,env),(lid,lrn),(vid,val),copy=learner_counts[lrn]>1)

class ChunkTasks(Filter[Iterable[Task], Iterable[Sequence[Task]]]):

    def __init__(self, max_tasks: int = None) -> None:
        self._max_tasks = max_tasks or None

    def filter(self, items: Iterable[Task]) -> Iterable[Sequence[Task]]:
        return self._chunks(items)

    def _chunks(self, items: Iterable[Task]) -> Iterable[Sequence[Task]]:
        items  = list(items)
        chunks = defaultdict(list)

        tasks_sans_env = [t for t in items if not t.env ]
        tasks_with_env = [t for t in items if     t.env ]

        for task in tasks_with_env:
            chunks[self._get_last_chunk(task.env)].append(task)

        for task in tasks_sans_env:
            yield [task]

        for task in chunks.pop('not_chunked',[]):
            yield [task]

        chunks_sorter = lambda c: min([c.env_id for c in c])
        chunk_sorter  = lambda t: (t.env_id, t.lrn_id if t.lrn else -1)

        for chunk in sorted(chunks.values(), key=chunks_sorter):
            yield from self._max_chunker(sorted(chunk, key=chunk_sorter), self._max_tasks)

    def _get_last_chunk(self, env):
        if isinstance(env, SourceFilters):
            for pipe in reversed(list(env)):
                if isinstance(pipe, Chunk):
                    return pipe
        return 'not_chunked'

    def _max_chunker(self, chunk, max_tasks):
        chunk = iter(chunk)
        batch = list(islice(chunk,max_tasks))
        while batch != []:
            yield batch
            batch = list(islice(chunk,max_tasks))

class ProcessTasks(Filter[Iterable[Task], Iterable[Any]]):

    def filter(self, chunk: Iterable[Task]) -> Iterable[Any]:

        chunk = list(chunk)
        empty_envs = set()

        if not chunk: return

        if len(chunk) > 1:
            #We sort to make sure cached envs are grouped. Sorting means we can free envs from memory as we go.
            chunk = sorted(chunk, key=lambda item: self._env_ids(item)+self._lrn_ids(item), reverse=True)

        while chunk:
            try:
                task = chunk.pop()

                env_id,env = (task.env_id,task.env)
                lrn_id,lrn = (task.lrn_id,task.lrn)
                val_id,val = (task.val_id,task.val)

                if task.copy: lrn = deepcopy(lrn)

                if env and not lrn and not val:
                    with CobaContext.logger.time(f"Peeking at Environment {env_id}..."):
                        peek_first(env.read())

                    with CobaContext.logger.time(f"Recording Environment {env_id} parameters..."):
                        yield ["T1", env_id, SafeEnvironment(env).params]

                if lrn and not env and not val:
                    with CobaContext.logger.time(f"Recording Learner {lrn_id} parameters..."):
                        yield ["T2", lrn_id, SafeLearner(lrn).params]

                if val and not env and not lrn:
                    with CobaContext.logger.time(f"Recording Evaluator {val_id} parameters..."):
                        yield ["T3", val_id, SafeEvaluator(val).params]

                if env and lrn and val and env_id not in empty_envs:

                    with CobaContext.logger.time(f"Peeking at Environment {env_id}..."):
                        interactions = peek_first(env.read())[1]

                    if not interactions:
                        CobaContext.logger.log(f"Environment {env_id} has nothing to evaluate (this is likely due to having too few interactions).")
                        empty_envs.add(env_id)
                        continue

                    class dummy_env:
                        _env = env
                        @property
                        def params(self): return SafeEnvironment(dummy_env._env).params #pragma: no cover
                        def read(self): return BatchSafe(Finalize()).filter(interactions)

                    with CobaContext.logger.time(f"Evaluating Learner {lrn_id} on Environment {env_id}..."):
                        env = dummy_env()
                        yield ["T4", (env_id, lrn_id, val_id), list(SafeEvaluator(val).evaluate(env,lrn))]
                        if hasattr(lrn,'finish') and task.copy: lrn.finish()

            except Exception as e:
                CobaContext.logger.log(e)

    def _env_ids(self, item: Task):
        return (item.env_id if item.env else -1,)

    def _lrn_ids(self, item: Task):
        return (item.lrn_id if item.lrn else -1,)



from typing import Union

from coba.exceptions import CobaException

from coba.pipes.primitives import Filter, Source, Sink, SourceFilters, FiltersFilter, FiltersSink, Foreach
from coba.pipes.multiprocessing import AsyncableLine

class Pipes:

    @staticmethod
    def join(*pipes: Union[Source, Filter, Sink]) -> Union[Source, Filter, Sink, AsyncableLine]:
        """Join a sequence of pipes into a single pipe.

        Args:
            pipes: a sequence of pipes.

        Returns:
            A single pipe that is a composition of the given pipes. The type of pipe returned
            is determined by the sequence given. A sequence of Filters will return a Filter. A
            sequence that begins with a Source and is followed by Filters will return a Source.
            A sequence that starts with Filters and ends with a Sink will return a Sink. A
            sequence that begins with a Source and ends with a Sink will return a completed pipe.
        """
        if len(pipes) == 0:
            raise CobaException("No pipes were passed to join.")

        if len(pipes) == 1 and any(hasattr(pipes[0],attr) for attr in ['read','filter','write']):
            return pipes[0]

        first = pipes[0 ] if not isinstance(pipes[0 ], Foreach) else pipes[0 ]._pipe
        last  = pipes[-1] if not isinstance(pipes[-1], Foreach) else pipes[-1]._pipe

        if hasattr(first,'read') and hasattr(last,'write'):
            return AsyncableLine(*pipes)

        if hasattr(first,'read') and hasattr(last,'filter'):
            return SourceFilters(*pipes)

        if hasattr(first,'filter') and hasattr(last,'filter'):
            return FiltersFilter(*pipes)

        if hasattr(first,'filter') and hasattr(last,'write'):
            return FiltersSink(*pipes)

        raise CobaException("An unknown pipe was passed to join.")



import warnings
import importlib

from itertools import chain, islice
from collections import defaultdict
from typing import TypeVar, Iterable, Tuple, Union, Sequence

from coba.exceptions import CobaExit

def coba_exit(message:str):
    #we ignore warnings before exiting in order to make jupyter's output a little cleaner
    warnings.filterwarnings("ignore",message="To exit: use 'exit', 'quit', or Ctrl-D.")
    raise CobaExit(message) from None

class PackageChecker:

    @staticmethod
    def matplotlib(caller_name: str) -> None:
        """Raise ImportError with detailed error message if matplotlib is not installed.

        Functionality requiring matplotlib should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires matplotlib.
        """
        PackageChecker._handle_import_error(caller_name, "matplotlib.pyplot", "matplotlib")

    @staticmethod
    def vowpalwabbit(caller_name: str) -> None:
        """Raise ImportError with detailed error message if vowpalwabbit is not installed.

        Functionality requiring vowpalwabbit should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires vowpalwabbit.
        """

        PackageChecker._handle_import_error(caller_name, "vowpalwabbit")

    @staticmethod
    def pandas(caller_name: str) -> None:
        """Raise ImportError with detailed error message if pandas is not installed.

        Functionality requiring pandas should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires pandas.
        """
        PackageChecker._handle_import_error(caller_name, "pandas")

    @staticmethod
    def numpy(caller_name: str) -> None:
        """Raise ImportError with detailed error message if numpy is not installed.

        Functionality requiring numpy should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires numpy.
        """
        PackageChecker._handle_import_error(caller_name, "numpy")

    @staticmethod
    def scipy(caller_name: str) -> None:
        """Raise ImportError with detailed error message if scipy is not installed.

        Functionality requiring scipy should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires scipy.
        """
        PackageChecker._handle_import_error(caller_name, "scipy")

    @staticmethod
    def sklearn(caller_name: str) -> None:
        """Raise ImportError with detailed error message if sklearn is not installed.

        Functionality requiring sklearn should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires sklearn.
        """
        PackageChecker._handle_import_error(caller_name, "sklearn", "scikit-learn")

    @staticmethod
    def torch(caller_name: str) -> None:
        """Raise ImportError with detailed error message if torch is not installed.

        Functionality requiring torch should call this helper and then lazily import.

        Args:
            caller_name: The name of the caller that requires torch.
        """
        PackageChecker._handle_import_error(caller_name, "torch")

    def _handle_import_error(caller_name:str, module_name:str, pkg_name:str = None):
        """
        Remarks:
            This pattern borrows heavily from sklearn. As of 6/20/2020 sklearn code could be found
            at https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/__init__.py
        """
        try:
            importlib.import_module(module_name)
        except ImportError:
            pkg_name = pkg_name or module_name
            coba_exit(f"ERROR: {caller_name} requires the {pkg_name} package. You can install this package via `pip install {pkg_name}`.")

class KeyDefaultDict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError( key )
        else:
            value = self.default_factory(key)
            self[key] = value
            return value

_T = TypeVar("_T")
def peek_first(items: Iterable[_T], n:int=1) -> Tuple[Union[_T,Sequence[_T]], Iterable[_T]]:
    items = iter(items)
    first = list(islice(items,n))

    items = [] if not first else chain(first,items)
    first = None if not first else first[0] if n==1 else first

    return first, items



from abc import ABC, abstractmethod
from typing import Sequence, Optional, Mapping

from coba.exceptions import CobaException
from coba.primitives.semantic import Action, Batch

def argmax(actions: Sequence[Action], rewards: 'Reward') -> Action:
    max_r = -float('inf')
    max_a = 0
    for a in actions:
        r = rewards.eval(a)
        if r > max_r:
            max_a = a
            max_r = r

    return max_a

class Reward(ABC):
    __slots__ = ()

    @abstractmethod
    def eval(self, action: Action) -> float:
        ...

    def to_json(self) -> Mapping[str, str]:
        return {key.lstrip('_'): getattr(self, key, None) for key in self.__slots__}

class IPSReward(Reward):
    __slots__ = ('_reward','_action')

    def __init__(self, reward: float, action: Action, probability: Optional[float] = None) -> None:
        self._reward = reward/(probability or 1)
        self._action = action

    def eval(self, arg: Action) -> float:
        return self._reward if arg == self._action else 0

    def __reduce__(self):
        #this makes the pickle smaller
        return IPSReward, (self._reward, self._action, 1)
    
    def __eq__(self, o: object) -> bool:
        return isinstance(o,IPSReward) and o._reward == self._reward and o._action == self._action

class L1Reward(Reward):
    __slots__ = ('_label',)

    def __init__(self, action: float) -> None:
        self._label = action

    def eval(self, action: float) -> float:
        return action-self._label if self._label > action else self._label-action 

    def __reduce__(self):
        #this makes the pickle smaller
        return L1Reward, (self._label,)

    def __eq__(self, o: object) -> bool:
        return isinstance(o,L1Reward) and o._label == self._label

class HammingReward(Reward):
    __slots__ = ('_labels',)

    def __init__(self, labels: Sequence[Action]) -> None:
        self._labels = set(labels)

    def eval(self, arg: Sequence[Action]) -> float:
        return len(self._labels.intersection(arg))/len(self._labels.union(arg))

    def __reduce__(self):
        #this makes the pickle smaller
        return HammingReward, (tuple(self._labels),)

class BinaryReward(Reward):
    __slots__ = ('_maxarg',)

    def __init__(self, action: Action) -> None:
        self._maxarg = action

    def eval(self, action: Action) -> float:
        return float(self._maxarg==action)

    def __reduce__(self):
        #this makes the pickle smaller
        return BinaryReward, (self._maxarg,)
    
    def __eq__(self, o: object) -> bool:
        return o == self._maxarg or (isinstance(o,BinaryReward) and o._maxarg == self._maxarg)

class SequenceReward(Reward):
    __slots__ = ('_actions','_rewards')

    def __init__(self, actions: Sequence[Action], rewards: Sequence[float]) -> None:
        if len(actions) != len(rewards): 
            raise CobaException("The given actions and rewards did not line up.")

        self._actions = actions
        self._rewards = rewards

    def eval(self, action: Action) -> float:
        return self._rewards[self._actions.index(action)]

    def __reduce__(self):
        #this makes the pickle smaller
        return SequenceReward, (self._actions,self._rewards)

    def __eq__(self, o: object) -> bool:
        return o == self._rewards or (isinstance(o,SequenceReward) and o._actions == self._actions and o._rewards == self._rewards)

class MappingReward(Reward):
    __slots__ = ('_mapping',)

    def __init__(self, mapping: Mapping[Action,float]) -> None:
        self._mapping = mapping

    def eval(self, action: Action) -> float:
        return self._mapping[action]

    def __reduce__(self):
        #this makes the pickle smaller
        return MappingReward, (self._mapping,)

    def __eq__(self, o: object) -> bool:
        return o == self._mapping or (isinstance(o,MappingReward) and o._mapping == self._mapping)

class MulticlassReward(Reward):
    __slots__=('_label',)
    
    def __init__(self, label: Action) -> None:
        self._label = label

    def eval(self, action: Action) -> float:
        return int(self._label == action)

    def __reduce__(self):
        return MulticlassReward, (self._label,)
    
    def __eq__(self, o: object) -> bool:
        return o == self._label or (isinstance(o,MulticlassReward) and o._label == self._label)

class BatchReward(Batch):
    def eval(self, actions: Sequence[Action]) -> Sequence[float]:
        return list(map(lambda r,a: r.eval(a), self, actions))



from coba.learners.primitives import Learner

class MisguidedLearner(Learner):
    """A contextual bandit learner that is given incorrect reward information. It is
    useful when generating logged data that is difficult in off-policy experiments.
    """
    def __init__(self, learner: Learner, shifter:float, scaler:float) -> None:
        self._learner = learner
        self._shifter = shifter
        self._scaler  = scaler

    @property
    def params(self):
        return {**self._learner.params, 'misguided': [self._shifter,self._scaler]}

    def request(self, context, actions, request):
        return self._learner.request(context, actions, request)

    def predict(self, context, actions):
        return self._learner.predict(context, actions)

    def learn(self, context, action, reward, probability,**kwargs):
        self._learner.learn(context, action, self._shifter + self._scaler*reward, probability,**kwargs)



from coba.random import CobaRandom
from coba.context import CobaContext, NullLogger

from coba.environments import Environments, ArffSource, CsvSource, LibSvmSource, ManikSource
from coba.environments import Interaction, SimulatedInteraction, LoggedInteraction, GroundedInteraction, LambdaSimulation

from coba.learners import Learner
from coba.learners import SafeLearner, FixedLearner, RandomLearner
from coba.learners import EpsilonBanditLearner, UcbBanditLearner
from coba.learners import CorralLearner, LinUCBLearner
from coba.learners import VowpalLearner, VowpalEpsilonLearner, VowpalSoftmaxLearner, VowpalBagLearner, VowpalRndLearner
from coba.learners import VowpalCoverLearner, VowpalRegcbLearner, VowpalSquarecbLearner, VowpalOffPolicyLearner
from coba.learners import MisguidedLearner

from coba.evaluators import OnPolicyEvaluator, OffPolicyEvaluator, ClassMetaEvaluator, ExplorationEvaluator

from coba.experiments  import Experiment, Result

from coba.encodings import InteractionsEncoder
from coba.utilities import peek_first
from coba.exceptions import CobaException

from coba.primitives.semantic import Batch
from coba.primitives.rewards import L1Reward, HammingReward, BinaryReward, IPSReward
from coba.primitives.rewards import SequenceReward, MappingReward, MulticlassReward, BatchReward

from coba.statistics import BootstrapCI, mean

__version__ = "7.1.4"



import math
import collections
import collections.abc
import warnings

from itertools import combinations
from operator import __mul__
from statistics import mean
from typing import Any, Sequence, Mapping, Hashable, Optional

from coba.encodings import InteractionsEncoder
from coba.learners import Learner
from coba.environments import Environment
from coba.exceptions import CobaExit
from coba.primitives import argmax
from coba.statistics import percentile
from coba.utilities import PackageChecker, peek_first

from coba.evaluators.primitives import Evaluator

class ClassMetaEvaluator(Evaluator):
    """Describe an Environment made from a Classification dataset.

    This evlaluator calculates a number of statistics which describe a given environment. These 
    statistics can be used to analyze what kinds of environments a given learner performs best on.
    To make the most of this evaluator sklearn should be installed.
    """

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Mapping[Any,Any]:

        #sources:
        #[1]: https://arxiv.org/pdf/1808.03591.pdf (lorena2019complex)
        #[2]: https://link.springer.com/content/pdf/10.1007/978-3-540-31883-5.pdf#page=468 (castiello2005meta)
        #[3]: https://link.springer.com/content/pdf/10.1007/s10044-012-0280-z.pdf (reif2014automatic)
        #[4]: https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.440.6255&rep=rep1&type=pdf

        #[3] found that information theoretic measures and landmarking measures are most important

        interactions = peek_first(environment.read())[1]

        if not interactions: return {}

        contexts, actions, rewards = zip(*[ (i['context'], i['actions'], i['rewards']) for i in interactions ])
        env_stats = {}

        X = [ InteractionsEncoder('x').encode(x=c) for c   in contexts             ]
        Y = [ argmax(a,r)                          for a,r in zip(actions,rewards) ]
        X = self._dense(X)

        classes = list(set(Y))
        feats   = list(range(len(X[0])))

        n = len(X)
        m = len(feats)
        k = len(classes)

        X_bin = self._bin(X,n/10)
        X_bin_by_f = list(zip(*X_bin))

        entropy_Y = self._entropy(Y)
        entropy_X = [self._entropy(x) for x in X_bin_by_f]

        mutual_XY_infos  = sorted([self._mutual_info(x,Y) for x in X_bin_by_f], reverse=True)

        #Information-Theoretic Meta-features
        env_stats["class_count"          ] = k
        env_stats["class_entropy"        ] = entropy_Y                # [1]
        env_stats["class_entropy_N"      ] = self._entropy_normed(Y)  # [1,2,3]
        env_stats["class_imbalance_ratio"] = self._imbalance_ratio(Y) # [1]

        env_stats["feature_numeric_count"  ] = sum([int(len(set(X_bin_by_f[f]))!=2) for f in feats])
        env_stats["feature_onehot_count"   ] = sum([int(len(set(X_bin_by_f[f]))==2) for f in feats])
        env_stats["feature_entropy_mean"   ] = mean([self._entropy(X_bin_by_f[f]) for f in feats])
        env_stats["feature_entropy_mean_N" ] = mean([self._entropy_normed(X_bin_by_f[f]) for f in feats])
        env_stats["joint_XY_entropy_mean"  ] = mean([self._entropy(list(zip(X_bin_by_f[f],Y))) for f in feats]) #[2,3]
        env_stats["joint_XY_entropy_mean_N"] = mean([self._entropy_normed(list(zip(X_bin_by_f[f],Y))) for f in feats])
        env_stats["mutual_XY_info_mean"    ] = mean(mutual_XY_infos) #[2,3]
        env_stats["mutual_XY_info_mean_N"  ] = mean(mutual_XY_infos)/entropy_Y if entropy_Y else None #[2,3]

        env_stats["mutual_XY_info_rank1"   ] = mutual_XY_infos[0]
        env_stats["mutual_XY_info_rank2"   ] = mutual_XY_infos[1] if len(mutual_XY_infos) > 1 else None
        env_stats["equivalent_num_X_attr"  ] = entropy_Y/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2,3]
        env_stats["noise_signal_ratio"     ] = (mean(entropy_X)-mean(mutual_XY_infos))/mean(mutual_XY_infos) if mean(mutual_XY_infos) else None #[2]

        env_stats["max_fisher_discrim"    ] = self._max_fisher_discriminant_ratio(X, Y) #[1]
        #env_stats["max_fisher_discrim_dir"] = self._max_directional_fisher_discriminant_ratio(X, Y) #[1] (this dies on large feature)
        env_stats["volume_overlapping"    ] = self._volume_overlapping_region(X,Y) #[1]
        env_stats["max_single_feature_eff"] = self._max_individual_feature_efficiency(X,Y) #[1]

        #Sparsity/Dimensionality measures [1,2,3]
        env_stats["feature_count"       ] = m
        env_stats["percent_nonzero_feat"] = mean([len(list(filter(None,x)))/len(x) for x in X])

        try:

            PackageChecker.sklearn("ClassEnvironmentTask")

            import numpy as np

            from sklearn.decomposition import TruncatedSVD
            from sklearn.neighbors import KNeighborsClassifier
            from sklearn.naive_bayes import GaussianNB
            from sklearn.metrics import f1_score, accuracy_score
            from sklearn.model_selection import cross_validate
            from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
            from sklearn.ensemble import RandomForestClassifier
            from sklearn.tree import DecisionTreeClassifier

            import sklearn.exceptions
            warnings.filterwarnings("ignore", category=sklearn.exceptions.FitFailedWarning)

            np_X = np.array(X)
            np_Y = np.array(Y)

            try:
                #1NN OOB [3,4]
                oob = np_Y[KNeighborsClassifier(n_neighbors=1).fit(np_X,np_Y).kneighbors(np_X, n_neighbors=2, return_distance=False)[:,1]]
                env_stats["1nn_accuracy"] = float(accuracy_score(np_Y,oob))
                env_stats["1nn_f1_macro"] = float(f1_score(np_Y,oob, average='macro'))
            except: #pragma: no cover
                pass

            try:
                #LDA [3,4]
                scr = cross_validate(LinearDiscriminantAnalysis(), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["lda_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["lda_f1_weighted"] = float(mean(scr['test_f1_weighted']))
            except: #pragma: no cover
                pass

            try:
                #Naive Bayes [3,4]
                scr = cross_validate(GaussianNB(), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["naive_bayes_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["naive_bayes_f1_weighted"] = float(mean(scr['test_f1_macro']))
            except: #pragma: no cover
                pass

            try:
                #Average Node Learner [3,4]
                scr = cross_validate(RandomForestClassifier(n_estimators=50,criterion='entropy',max_depth=1), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["average_node_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["average_node_f1_macro"] = float(mean(scr['test_f1_weighted']))
            except: #pragma: no cover
                pass

            try:
                #Best Node Learner [3,4]
                scr = cross_validate(DecisionTreeClassifier(criterion='entropy',max_depth=1), np_X, np_Y, scoring=('accuracy','f1_macro'))
                env_stats["best_node_accuracy"] = float(mean(scr['test_accuracy']))
                env_stats["best_node_f1_macro"] = float(mean(scr['test_f1_macro']))
            except: #pragma: no cover
                pass

            try:
                #pca effective dimensions [1]
                centered_x= np_X - np_X.mean(axis=0)
                pca_var = TruncatedSVD(n_components=min(np_X.shape[1]-1,1000)).fit(centered_x).explained_variance_ratio_
                env_stats["pca_top_1_pct"] = pca_var[0:1].sum()
                env_stats["pca_top_2_pct"] = pca_var[0:2].sum()
                env_stats["pca_top_3_pct"] = pca_var[0:3].sum()
                env_stats["pca_dims_95"  ] = int(sum(np.cumsum(pca_var)<.95)+1)
            except: #pragma: no cover
                pass

            #sklearn's CCA doesn't seem to work with sparse so I'm leaving it out for now depsite [3]

        except CobaExit:
            pass

        return env_stats

    def _entropy(self, items: Sequence[Hashable]) -> float:
        return -sum([count/len(items)*math.log2(count/len(items)) for count in collections.Counter(items).values()])

    def _entropy_normed(self, items: Sequence[Hashable]) -> float:
        return self._entropy(items)/(math.log2(len(set(items))) or 1)

    def _mutual_info(self, items1: Sequence[Hashable], items2: Sequence[Hashable]) -> float:
        return self._entropy(items1) + self._entropy(items2) - self._entropy(list(zip(items1,items2)))

    def _imbalance_ratio(self, items: list) -> float:
        #Equation (37) and (38) in [1]

        counts = collections.Counter(items).values()
        n      = len(items)
        k      = len(counts)

        if n == 0: return None
        if max(counts) == n: return 1

        IR = (k-1)/k*sum([c/(n-c) for c in counts])
        return 1 - 1/IR

    def _max_fisher_discriminant_ratio(self, X, Y) -> float:
        #Equation (3) in [1]
        #needs more testing

        feats    = list(range(len(X[0])))
        Y_set    = set(Y)
        Y_counts = collections.Counter(Y)

        X_by_f   = collections.defaultdict(list)
        X_by_fy  = collections.defaultdict(lambda: collections.defaultdict(list))

        for x,y in zip(X,Y):
            for f in feats:
                xf = x[f]
                X_by_fy[f][y].append(xf)
                X_by_f[f].append(xf)

        mean_f  = { f:      mean(X_by_f[f])                      for f in feats}
        mean_fy = { f: { y: mean(X_by_fy[f][y]) for y in Y_set } for f in feats}

        max_ratio = 0

        for f in feats:
            ratio_numer = sum([Y_counts[y]*(mean_fy[f][y]-mean_f[f])**2 for y in Y_set])
            ratio_denom = sum([(X_by_fy[f][y][i]-mean_fy[f][y])**2 for y in Y_set for i in range(Y_counts[y]) ])
            if ratio_denom != 0:
                max_ratio   = max(max_ratio,ratio_numer/ratio_denom)

        return 1/(1+max_ratio)

    def _max_directional_fisher_discriminant_ratio(self, X, Y) -> float:
        #equation (4) in [1]
        #this code is currently not used because it can take an incredibly
        #long time to calculate due to np.linalg.inv so we "no cover" it

        try:
            PackageChecker.sklearn('')

            import numpy as np
            from sklearn.covariance import shrunk_covariance

            Y_set = set(Y)
            X     = np.array(X).T #transpose so that equations align with paper
            Y     = np.array(Y)

            X_by_y = { y: X[:,Y==y]  for y in Y_set }
            OVO    = []

            for y1,y2 in combinations(Y_set,2):
                mu_c1 = X_by_y[y1].mean(axis=1).reshape(-1,1)
                mu_c2 = X_by_y[y2].mean(axis=1).reshape(-1,1)

                p1 = X_by_y[y1].shape[1]/(X_by_y[y1].shape[1]+X_by_y[y2].shape[1])
                p2 = X_by_y[y2].shape[1]/(X_by_y[y1].shape[1]+X_by_y[y2].shape[1])

                s1 = (X_by_y[y1]-mu_c1) @ (X_by_y[y1]-mu_c1).T
                s2 = (X_by_y[y2]-mu_c2) @ (X_by_y[y2]-mu_c2).T
                B  = (mu_c1-mu_c2) @ (mu_c1-mu_c2).T

                s1 = shrunk_covariance(s1)
                s2 = shrunk_covariance(s2)
                B  = shrunk_covariance(B)

                W = p1*s1+p2*s2
                d = np.linalg.inv(W) @ (mu_c1-mu_c2)

                OVO.append( ((d.T@B@d) / (d.T@W@d))[0,0] )

            return 1/(1+mean(OVO)) if OVO else None

        except np.linalg.LinAlgError:#pragma: no cover
            return None
        except CobaExit:#pragma: no cover
            return None

    def _volume_overlapping_region(self, X, Y) -> float:
        #equation (9) in [1]

        X_by_y = collections.defaultdict(list)
        for x,y in zip(X,Y): X_by_y[y].append(x)
        F_by_y = { y: list(zip(*x)) for y,x in X_by_y.items()}

        minmax = lambda f,y1,y2: min(max(F_by_y[y1][f]), max(F_by_y[y2][f]))
        maxmin = lambda f,y1,y2: max(min(F_by_y[y1][f]), min(F_by_y[y2][f]))
        maxmax = lambda f,y1,y2: max(max(F_by_y[y1][f]), max(F_by_y[y2][f]))
        minmin = lambda f,y1,y2: min(min(F_by_y[y1][f]), min(F_by_y[y2][f]))

        OVO = []

        for y1,y2 in combinations(set(Y),2):
            F2 = 1
            for f in range(len(X[0])):
                if maxmax(f,y1,y2)-minmin(f,y1,y2) != 0:
                    F2 *= max(0, (minmax(f,y1,y2)-maxmin(f,y1,y2)))/(maxmax(f,y1,y2)-minmin(f,y1,y2))

            OVO.append(F2)

        return mean(OVO) if OVO else None

    def _max_individual_feature_efficiency(self, X, Y) -> float:
        #equation (11) in [1]

        try:
            PackageChecker.numpy('ClassEnvironmentTask')
            import numpy as np

            X_by_f = np.array(list(zip(*X)))
            X_by_y = collections.defaultdict(list)
            for x,y in zip(X,Y): X_by_y[y].append(x)
            F_by_y  = { y: list(zip(*x)) for y,x in X_by_y.items()}
            L_by_fy = { f: { y:percentile(F_by_y[y][f],[0.0,1.0])  for y in F_by_y } for f in range(len(X[0])) }

            minmax = lambda f,y1,y2: min(L_by_fy[f][y1][1], L_by_fy[f][y2][1])
            maxmin = lambda f,y1,y2: max(L_by_fy[f][y1][0], L_by_fy[f][y2][0])

            OVO    = []

            for y1,y2 in combinations(set(Y),2):

                n_o = []
                for f in range(len(X[0])):
                    n_o.append(int(((X_by_f[f] <= minmax(f,y1,y2)) & (maxmin(f,y1,y2) <= X_by_f[f])).sum()))

                OVO.append(min(n_o)/len(X))

            return mean(OVO) if OVO else None

        except CobaExit:
            return None

    def _dense(self, X) -> Sequence[Sequence[float]]:

        if not isinstance(X[0],dict):
            return X
        else:
            #Convert the sparse dicts into a compact dense array.
            #This is required by for a number of the analysis in this task.
            feats = sorted(set().union(*X)) #sort to make unit tests repeatable
            dense_X = [ [0]*len(feats) for _ in range(len(X)) ]

            for i,x in enumerate(X):
                for j,f in enumerate(feats):
                        dense_X[i][j] = x.get(f,0)

            return dense_X

    def _bin(self, X: Sequence[Sequence[float]], n_bins:int, lower:float=0.05, upper:float=0.95) -> Sequence[Sequence[float]]:
        X_by_f = list(zip(*X))
        lim_f  = { f: percentile(X_by_f[f],[lower,upper]) for f in range(len(X_by_f)) }
        X_bin  = [ [ round((n_bins-1)*(x[f]-lim_f[f][0])/((lim_f[f][1]-lim_f[f][0]) or 1)) for f in range(len(X_by_f)) ] for x in X]
        return X_bin



import pickle

from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED
from itertools import islice, repeat, chain
from typing import Union, Sequence, Mapping, Iterable, Any

from coba.context import CobaContext
from coba.pipes import Source, Sink, Filter
from coba.environments.primitives import Environment, Interaction

class ObjectsToZipMember(Sink[Iterable[Sequence[object]]]):
    def __init__(self, zip:str):
        self._zip = zip
        self._start = 0
        if Path(self._zip).exists():
            for n in ZipFile(self._zip).namelist():
                try:
                    self._start = max(self._start,int(n)+1)
                except ValueError: #pragma: no cover
                    pass

    def write(self, envs: Iterable[Sequence[object]]) -> None:
        for i,env in enumerate(envs,self._start):
            with CobaContext.logger.time(f"Writing environment {i}..."):
                # Dump before opening the file so we
                # can keep it open for as short a time 
                # as possible.
                dumps = list(map(pickle.dumps,env))
                with ZipFile(self._zip,mode='a',compression=ZIP_DEFLATED) as zip:
                    with zip.open(str(i),mode='w') as f:
                        f.writelines(dumps)

class ZipMemberToObjects(Source[Iterable[object]]):
    def __init__(self, zip:str, member:str):
        self._zip    = zip
        self._member = member

    def read(self) -> Iterable[object]:
        try:
            with ZipFile(self._zip) as z:
                with z.open(self._member) as m:
                    yield from map(pickle.load, repeat(m))
        except EOFError:
            pass

class EnvironmentsToObjects(Filter[Environment, Iterable]):
    def filter(self, env: Union[Environment,Sequence[Environment]]) -> Iterable[Iterable[object]]:        
        with CobaContext.logger.time("Materializing environment..."):
            yield list(self._env_to_objects(env))

    def _env_to_objects(self,env):
        from coba import __version__ #imported here to avoid circular dependency
        yield {"version":2,"coba_version":__version__}
        yield env.params
        I = iter(env.read())
        batch = list(islice(I,1000))
        while batch: 
            yield batch
            batch = list(islice(I,1000))

class EnvironmentFromObjects(Environment):
    def __init__(self, source: Source[Iterable[object]]) -> None:
        self._source = source

    @property
    def params(self) -> Mapping[str,Any]:
        return list(islice(self._source.read(),1,2))[0]

    def read(self) -> Iterable[Interaction]:

        I = iter(self._source.read())
        version_data = list(islice(I,2))[0]
        
        if version_data['version'] == 1: #pragma: no cover
            yield from I
        else:
            yield from chain.from_iterable(I)



from abc import ABC, abstractmethod
from typing import Any, Mapping, Iterable, Union, Optional, Callable

from coba.learners import Learner
from coba.environments import Environment

class Evaluator(ABC):

    @property
    def params(self) -> Mapping[str,Any]:
        """Parameters describing the learner (used for descriptive purposes only).

        Remarks:
            These will become columns in the learners table of experiment results.
        """
        return {}

    @abstractmethod
    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Union[Mapping[Any,Any],Iterable[Mapping[Any,Any]]]:
        """Evaluate the learner on the given interactions.

        Args:
            environment: The Environment we want to evaluate against.
            learner: The Learner that we wish to evaluate.

        Returns:
            Evaluation results
        """
        ...

class SafeEvaluator(Evaluator):
    def __init__(self, evaluator: Union[Evaluator, Callable[[Environment,Learner], Union[Mapping, Iterable[Mapping]]]]) -> None:
        self.evaluator = evaluator if not isinstance(evaluator, SafeEvaluator) else evaluator.evaluator

    @property
    def params(self):
        try:
            params = self.evaluator.params
        except:
            params = {}

        if type(self.evaluator).__name__ == 'function':
            params['eval_type'] = self.evaluator.__name__
        else:
            params['eval_type'] = type(self.evaluator).__name__

        return params

    def evaluate(self, environment: Optional[Environment], learner: Optional[Learner]) -> Union[Mapping, Iterable[Mapping]]:
        if callable(self.evaluator):
            return self.evaluator(environment,learner)
        else:
            return self.evaluator.evaluate(environment,learner)

def get_ope_loss(learner) -> float:
    # OPE loss metric is only available for VW models
    try:
        return learner.learner._vw._vw.get_sum_loss()
    except AttributeError:
        return float("nan")
